"""
basepair.py

Define base pairs in a DNA chain
"""

from collections import OrderedDict
from copy import deepcopy

import numpy as np
import pandas as pd

from fractaldna.dna_models import dnapositions as dpos
from fractaldna.dna_models import molecules
from fractaldna.utils.rotations import eulerMatrix

# Positions in Angstroms
PHOSPHATE_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.PHOSPHATE).find_center()
SUGAR_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.DEOXYRIBOSE).find_center()
GUANINE_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.GUANINE).find_center()
ADENINE_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.ADENINE).find_center()
THYMINE_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.THYMINE).find_center()
CYTOSINE_POS = dpos.MoleculeFromAtoms.from_cylindrical(dpos.CYTOSINE).find_center()


PHOSPHATE_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.PHOSPHATE, inverse=True
).find_center()
SUGAR_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.DEOXYRIBOSE, inverse=True
).find_center()
GUANINE_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.GUANINE, inverse=True
).find_center()
ADENINE_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.ADENINE, inverse=True
).find_center()
THYMINE_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.THYMINE, inverse=True
).find_center()
CYTOSINE_POS_OPP = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.CYTOSINE, inverse=True
).find_center()


class BasePair:
    """
    Defines a base pair

    BasePair(base, chain=-1, position=np.zeros(3), rotation=np.zeros(3))

    Create a base pair defined by a particular base GTAC (in the 5' dirn)

    :param base:   defining base (GATC)
    :param chain:    int to indicate the id of the current DNA macromolecule
    :param position: position of base pair relative to (0, 0, 0)
    :param rotation: rotation euler angles of base pair relative to local xyz
    :param index:    index of base pair in chain
    """

    pairings = {
        "G": [molecules.Guanine, molecules.Cytosine],
        "A": [molecules.Adenine, molecules.Thymine],
        "T": [molecules.Thymine, molecules.Adenine],
        "C": [molecules.Cytosine, molecules.Guanine],
    }
    moleculeNames = [
        "leftPhosphate",
        "leftSugar",
        "leftBasePair",
        "rightPhosphate",
        "rightSugar",
        "rightBasePair",
    ]

    def __init__(
        self,
        base: str,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """Constructor"""
        if base not in ["G", "A", "T", "C"]:
            raise ValueError("base must be either G, A, T, C")

        # start building at 0, so we can rotate in a neutral frame
        # before translating
        self.position = np.zeros(3)
        self.rmatrix = np.eye(3)
        self.chain = chain
        position = np.array(deepcopy(position), dtype=float)
        rotation = np.array(deepcopy(rotation), dtype=float)

        leftPhosphateRot = np.zeros(3)
        leftPhosphatePos = deepcopy(PHOSPHATE_POS)
        self.leftPhosphate = molecules.Triphosphate(
            strand=0,
            chain=chain,
            position=leftPhosphatePos,
            rotation=leftPhosphateRot,
            index=index,
        )
        leftSugarRot = np.zeros(3)
        leftSugarPos = deepcopy(SUGAR_POS)
        self.leftSugar = molecules.DNASugar(
            strand=0,
            chain=chain,
            position=leftSugarPos,
            rotation=leftSugarRot,
            index=index,
        )

        rightPhosphateRot = np.zeros(3)
        rightPhosphatePos = deepcopy(PHOSPHATE_POS_OPP)
        self.rightPhosphate = molecules.Triphosphate(
            strand=1,
            chain=chain,
            position=rightPhosphatePos,
            rotation=rightPhosphateRot,
            index=index,
        )

        rightSugarRot = np.zeros(3)
        rightSugarPos = deepcopy(SUGAR_POS_OPP)
        self.rightSugar = molecules.DNASugar(
            strand=1,
            chain=chain,
            position=rightSugarPos,
            rotation=rightSugarRot,
            index=index,
        )

        if base == "G":
            leftpos = deepcopy(GUANINE_POS)
            rightpos = deepcopy(CYTOSINE_POS_OPP)
        elif base == "C":
            leftpos = deepcopy(CYTOSINE_POS)
            rightpos = deepcopy(GUANINE_POS_OPP)
        elif base == "T":
            leftpos = deepcopy(ADENINE_POS)
            rightpos = deepcopy(THYMINE_POS_OPP)
        elif base == "A":
            leftpos = deepcopy(THYMINE_POS)
            rightpos = deepcopy(ADENINE_POS_OPP)

        leftBasePairRot = np.zeros(3)
        leftBasePairPos = leftpos

        rightBasePairRot = np.zeros(3)
        rightBasePairPos = rightpos

        bases = self.pairings[base]
        self.leftBasePair = bases[0](
            strand=0,
            chain=chain,
            position=leftBasePairPos,
            rotation=leftBasePairRot,
            index=index,
        )
        self.rightBasePair = bases[1](
            strand=1,
            chain=chain,
            position=rightBasePairPos,
            rotation=rightBasePairRot,
            index=index,
        )
        self.molecules = [
            self.leftPhosphate,
            self.leftSugar,
            self.leftBasePair,
            self.rightPhosphate,
            self.rightSugar,
            self.rightBasePair,
        ]

        self.moleculeDict = OrderedDict(zip(self.moleculeNames, self.molecules))

        self.rotate(rotation, about_origin=False)
        self.translate(position)  # sets self.position

    def translate(self, translation: np.array) -> None:
        """
        BasePair.translate(translation)

        translate the base pair by the specified 3-vector array
        """
        for molecule in self.moleculeDict.values():
            molecule.translate(translation)
        self.position = self.position + np.array(translation)
        return None

    def rotate(self, rotation: np.array, about_origin: bool = False) -> None:
        """
        BasePair.rotation(rotation, about_origin=False)

        Rotate elements in base pair

        If about_origin is True, the rotation is with respect to (0, 0, 0)
        If about_origin is False, the rotation is with respect to the base pair position
        """
        if rotation.size == 3:
            rmatrix = eulerMatrix(rotation[0], rotation[1], rotation[2])
        elif rotation.shape == (3, 3):
            rmatrix = rotation
        else:
            return NotImplementedError("The rotation was invalid")

        self.rmatrix = np.dot(rmatrix, self.rmatrix)
        if about_origin is True:
            self.position = np.dot(rmatrix, self.position)

        for molecule in self.moleculeDict.values():
            if about_origin is True:
                molecule.position = np.dot(rmatrix, molecule.position)
            else:
                local_pos = molecule.position - self.position
                molecule.position = np.dot(rmatrix, local_pos) + self.position
            molecule.rotate(rotation)

        return None

    def getCenter(self) -> np.array:
        """
        basepair.getCenter()

        Get the geometric center of the base-pair
        (non-weighted average molecule position)
        """
        positions = [mol.position for mol in self.moleculeDict.values()]
        return np.mean(np.array(positions), axis=1)

    def setNewChain(self, chainIdx: int) -> None:
        """
        BasePair.setNewChain(chainIdx)

        Reset the chain index for all molecules and the base pair
        """
        self.chain = chainIdx
        for molecule in self.moleculeDict.values():
            molecule.chain = chainIdx

        return None

    def set_bp_index(self, bp_idx: int) -> None:
        """
        BasePair.setNewChain(chainIdx)

        Reset the chain index for all molecules
        """
        for molecule in self.moleculeDict.values():
            molecule.index = bp_idx

        return None

    def to_text(self, seperator: str = " ") -> str:
        """
        BasePair.to_text(self, seperator=" ")
        Return a description of the molecules in the base pair as text
        """
        output = []
        for molecule in self.moleculeDict.values():
            output.append(molecule.to_text(seperator=seperator))

        return "".join(output)

    def to_frame(self) -> pd.DataFrame:
        """Return base pair as data frame of molecules"""
        return pd.DataFrame([mol.to_series() for mol in self.moleculeDict.values()])

    def __getitem__(self, key: str) -> molecules.Molecule:
        return self.moleculeDict[key]

    def iterMolecules(self):
        return self.moleculeDict.items()



"""
molecules.py

Define the molecules that make up a simple DNA chain
"""

from copy import deepcopy

import numpy as np
import pandas as pd

from fractaldna.dna_models import dnapositions as dpos
from fractaldna.utils import rotations as rot

# Physical parameters, all in Angstrom
GUANINE_SIZE = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.GUANINE
).find_equivalent_half_lengths()
ADENINE_SIZE = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.ADENINE
).find_equivalent_half_lengths()
THYMINE_SIZE = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.THYMINE
).find_equivalent_half_lengths()
CYTOSINE_SIZE = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.CYTOSINE
).find_equivalent_half_lengths()

SUGAR_RADIUS = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.DEOXYRIBOSE
).find_equivalent_radius()
PHOSPHATE_RADIUS = dpos.MoleculeFromAtoms.from_cylindrical(
    dpos.PHOSPHATE
).find_equivalent_radius()


class Molecule:
    """Create a molecule

    :param name: molecule name
    :param shape: shape of molecule
    :param dimensions: dimensions of molecule shape
        (3-vector for xyz-dims, int/float for radius)
    :param strand: int to hold an ID related to a localised strand
    :param chain: int to hold an ID related to a macromolecule/protein or
        DNA chain
    :param position: 3-vector for molecule position relative to global axis
    :param rotation: 3-vector of euler angles (radians) for molecule
        rotation relative to the global xyz axis
    :param index: index of base pair along chain
    """

    def __init__(
        self,
        name: str,
        shape: str,
        dimensions: np.array,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        if type(dimensions) in [int, float, np.float64]:
            self.dimensions = np.array([deepcopy(dimensions)] * 3)
        else:
            assert len(dimensions) == 3, "Position is invalid"
            self.dimensions = deepcopy(dimensions)

        self.shape = deepcopy(shape)
        self.name = deepcopy(name)
        self.position = deepcopy(position)
        self.rotation = deepcopy(rotation)
        self.strand = strand
        self.chain = chain
        self.index = index

    def translate(self, translation: np.array) -> None:
        """
        Molecule.translate(translation)

        :param translation: Translate the molecule by (x, y, z)
        """
        self.position = self.position + translation
        return None

    def rotate(self, rotation: np.array) -> None:
        """
        Molecule.rotate(rotation)

        Rotate molecule by [X_angle, Y_angle, Z_angle]

        :param rotation: Euler angles for rotation
        """
        if rotation.size == 3:
            rmatrix = rot.eulerMatrix(rotation[0], rotation[1], rotation[2])
        elif rotation.shape == (3, 3):
            rmatrix = rotation
        else:
            return NotImplementedError("The rotation was invalid")
        # assert rotation.size == 3, "Rotation array must be length-3"
        oldrotation = rot.eulerMatrix(*self.rotation)
        newrotation = rmatrix
        self.rotation = rot.getEulerAngles(np.dot(newrotation, oldrotation))
        return None

    def to_text(self, seperator: str = " ") -> str:
        """Return a text description of the molecule
        Molecule.toText(seperator=" ")

        :param seperator: seperation character
        """
        return (
            seperator.join(
                [
                    self.name,
                    self.shape,
                    str(self.chain),
                    str(self.strand),
                    str(self.index),
                    " ".join(map(str, self.dimensions)),
                    " ".join(map(str, self.position)),
                    " ".join(map(str, self.rotation)),
                ]
            )
            + "\n"
        )

    def to_series(self) -> pd.Series:
        """Convert molecule to a pandas series

        :return: Series representation of the molecule
        """
        return pd.Series(
            {
                "name": self.name,
                "shape": self.shape,
                "chain_idx": self.chain,
                "strand_idx": self.strand,
                "bp_idx": self.index,
                "size_x": self.dimensions[0],
                "size_y": self.dimensions[1],
                "size_z": self.dimensions[2],
                "pos_x": self.position[0],
                "pos_y": self.position[1],
                "pos_z": self.position[2],
                "rot_x": self.rotation[0],
                "rot_y": self.rotation[1],
                "rot_z": self.rotation[2],
            }
        )


# Define some standard molecules
class Guanine(Molecule):
    """
    Guanine molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        super().__init__(
            "Guanine",
            "ellipse",
            GUANINE_SIZE,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )


class Adenine(Molecule):
    """Adenine molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """Constructor"""
        super().__init__(
            "Adenine",
            "ellipse",
            ADENINE_SIZE,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )


class Thymine(Molecule):
    """Thymine molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        super().__init__(
            "Thymine",
            "ellipse",
            THYMINE_SIZE,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )


class Cytosine(Molecule):
    """
    Cytosine molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        super().__init__(
            "Cytosine",
            "ellipse",
            CYTOSINE_SIZE,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )


class DNASugar(Molecule):
    """
    DNASugar molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        super().__init__(
            "Sugar",
            "sphere",
            SUGAR_RADIUS,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )


class Triphosphate(Molecule):
    """
    Triphosphate molecule

    :param strand: strand ID
    :param chain: Chain ID
    :param position: position array (3-vector)
    :param rotation: rotation array (euler angles)
    :param index: base pait index
    """

    def __init__(
        self,
        strand: int = -1,
        chain: int = -1,
        position: np.array = np.zeros(3),
        rotation: np.array = np.zeros(3),
        index: int = 0,
    ):
        """constructor"""
        super().__init__(
            "Phosphate",
            "sphere",
            PHOSPHATE_RADIUS,
            strand=strand,
            chain=chain,
            position=position,
            rotation=rotation,
            index=index,
        )



"""
Generate a sequence of a GATC values of a certain length
"""

import numpy as np

from fractaldna.utils.constants import BP_SEPARATION


def get_dna_string(length, separation=BP_SEPARATION):
    """Return a string of letters representing a DNA strand of a set length
    get_dna_string(length, separation)

    args:
        length: length of strand to construct (default units angstrom)

    kwargs:
        separation: specified separation between base pairs, imported from
                    utils folder

    notes:
        Both length and separation should have the same units, the default
        separation is set in angstroms
    """
    candidates = ["G", "A", "T", "C"]
    strand = []
    for ii in range(0, int(length // separation)):
        strand.append(np.random.choice(candidates))

    return "".join(strand)



from typing import Callable, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from mpl_toolkits.mplot3d import Axes3D  # NOQA
from tqdm import tqdm

from . import hilbert

try:
    from mayavi import mlab

    maya_imported = True
except ImportError:
    maya_imported = False
    print("Could not import mayavi libraries, 3d plotting is disabled")
    print("MayaVi may need Python2")


def rot_ax_angle(axis: Union[List, np.array], angle: float) -> np.array:
    """Build the rotation matrix for a given rotation around an axis

    :param axis: Rotation axis (3-vector)
    :param angle: rotation angle (radians)

    :returns: The rotation matrix
    """
    ax = axis / np.sqrt(np.sum(np.array(axis) ** 2))
    ux = ax[0]
    uy = ax[1]
    uz = ax[2]

    c = np.cos(angle)
    s = np.sin(angle)

    xx = c + ux**2 * (1 - c)
    xy = ux * uy * (1 - c) - uz * s
    xz = ux * uz * (1 - c) + uy * s

    yx = uy * ux * (1 - c) + uz * s
    yy = c + uy**2 * (1 - c)
    yz = uy * uz * (1 - c) - ux * s

    zx = uz * ux * (1 - c) - uy * s
    zy = uz * uy * (1 - c) + ux * s
    zz = c + uz**2 * (1 - c)

    return np.array([[xx, xy, xz], [yx, yy, yz], [zx, zy, zz]])


def getEulerAngles(rotmatrix: np.array) -> Tuple[float, float, float]:
    """Get the euler angles from a rotation matrix

    :param rotmatrix: 3x3 rotation matrix
    :return: Euler psi angle
    """
    sintheta = rotmatrix[2, 0]
    if abs(sintheta) != 1:
        theta = -np.arcsin(rotmatrix[2, 0])
        costheta = np.cos(theta)
        psi = np.arctan2(rotmatrix[2, 1] / costheta, rotmatrix[2, 2] / costheta)
        phi = np.arctan2(rotmatrix[1, 0] / costheta, rotmatrix[0, 0] / costheta)
    else:
        phi = 0
        if sintheta < 0:  # Positive case
            theta = np.pi / 2.0
            psi = phi + np.arctan2(rotmatrix[0, 1], rotmatrix[0, 2])
        else:
            theta = -np.pi / 2.0
            psi = -phi + np.arctan2(-rotmatrix[0, 1], -rotmatrix[0, 2])

    return psi, theta, phi


class Voxel:
    """
    Position, rotation and form of a DNA placement voxel.

    The principal axes are used to determine if the DNA would
    undergo a 90˚ rotation as it passes through the voxel.

    One way to think of them is to consider that the path of your
    DNA or curve through the voxel is along a continuous z-axis,
    perpendicular to an X-Y plane. The Principal Axes could be the
    vectors that describe the direction of the X-axis.

    Class Notes:
    psi, theta, phi: Euler rotations about the X, Y and Z axes
    (XYZ rotation order)

    type: code corresponding to geometrical shape
    pos: XYZ position

    Voxel(pos, inHeading, inPrincipal, outHeading, outPrincipal)

    :param pos: XYZ-position of voxel
    :param inHeading: vector of the DNA heading "into" the voxel.
    :param inPrincipal: vector of the DNA's principal axis at the
        entrace to the voxel
    :param outHeading: vector of the DNA heading "out" of the voxel.
    :param outPrincipal: vector of the DNA's principal axis at the
        exit to the voxel
    """

    types = {"straight": 1, "straighttwist": 2, "turn": 3, "turntwist": 4}
    types_inverse = {v: k for (k, v) in types.items()}
    defaultPrincipal = np.array([1, 0, 0])
    defaultHeading = np.array([0, 0, 1])
    defaultOrtho = np.cross(defaultPrincipal, defaultHeading)
    defaultAxis = np.array([defaultPrincipal, -defaultOrtho, defaultHeading])
    defaultAxis = np.transpose(defaultAxis)
    defaultAxisInv = np.linalg.inv(defaultAxis)

    def __init__(
        self,
        pos: np.array,
        inHeading: np.array,
        inPrincipal: np.array,
        outHeading: np.array,
        outPrincipal: np.array,
    ):
        """
        Constructor
        """
        # Clean and vet input
        pos = np.around(pos, decimals=8)
        inHeading = np.around(inHeading, decimals=8)
        inPrincipal = np.around(inPrincipal, decimals=8)
        outHeading = np.around(outHeading, decimals=8)
        outPrincipal = np.around(outPrincipal, decimals=8)

        assert (inHeading != inPrincipal).any(), (
            "degenerate entry vectors: " + str(inHeading) + str(inPrincipal)
        )
        assert (outHeading != outPrincipal).any(), (
            "degenerate exit vectors: " + str(inPrincipal) + str(outPrincipal)
        )

        self.inHeading = inHeading / np.linalg.norm(inHeading)
        self.inPrincipal = inPrincipal / np.linalg.norm(inPrincipal)
        self.outHeading = outHeading / np.linalg.norm(outHeading)
        self.outPrincipal = outPrincipal / np.linalg.norm(outPrincipal)
        self.pos = pos

        sameHeading = (self.inHeading == self.outHeading).all()
        samePrincipal = (self.inPrincipal == self.outPrincipal).all() or (
            self.inPrincipal == -self.outPrincipal
        ).all()

        if sameHeading:
            if samePrincipal:
                self.type = self.types["straight"]
            else:
                self.type = self.types["straighttwist"]
        else:
            if samePrincipal:
                self.type = self.types["turn"]
            else:
                self.type = self.types["turntwist"]

        # Now we need to define the euler rotations of the pixel box
        # We assume that "No Rotation" involves the path entering with a +z
        # direction heading with the principal axis being the +x direction
        if self.type in [self.types["straight"], self.types["straighttwist"]]:
            paxis = self.inPrincipal
        else:
            paxis = self.outHeading

        self.orth = np.cross(paxis, self.inHeading)
        self.axis = np.array([paxis, -self.orth, self.inHeading])
        self.axis = np.transpose(self.axis)
        self.rotation = np.dot(self.axis, self.defaultAxisInv)

        # Euler angles
        # Using (x, y, z) rotations = (psi, theta, phi)
        self.psi, self.theta, self.phi = getEulerAngles(self.rotation)

    def to_text(self, sep: str = " ") -> str:
        """Print a textual representation of the voxel as
        KIND POS_X POS_Y POS_Z EUL_PSI EUL_THETA EUL_PHI
        """
        if len(sep) == 0:
            raise ValueError("Separator cannot be a zero-length string")
        l = (
            [self.types_inverse[self.type]]
            + list(map(str, list(self.pos)))
            + list(map(str, [self.psi, self.theta, self.phi]))
        )
        return sep.join(l)

    def to_series(self):
        """Return the voxel as a pandas series"""
        return pd.Series(
            {
                "KIND": self.types_inverse[self.type],
                "POS_X": self.pos[0],
                "POS_Y": self.pos[1],
                "POS_Z": self.pos[2],
                "EUL_PSI": self.psi,
                "EUL_THETA": self.theta,
                "EUL_PHI": self.phi,
            }
        )


class VoxelisedFractal:
    """
    Class containing a voxelised representation of a fractal

    Typically created using the VoxelisedFractal.fromSeed class method

    fractal = VoxelisedFractal.fromSeed('X', 1)
    fractal.fractal contains a list of 'Voxels' representing the DNA Path

    """

    def __init__(self):
        """Constructor"""
        self.fractal = []

    def __len__(self) -> int:
        return self.fractal.__len__()

    def to_text(self, sep: str = " ", comment: str = "#"):
        """Return a textual representation of the fractal as
            KIND POS_X POS_Y POS_Z EUL_PSI EUL_THETA EUL_PHI

        :param sep: string to use to separate fields (default space)
        :param comment: string to use to denote comment in first line
            (default hash: #)
        """
        if len(sep) == 0:
            raise ValueError("Separator cannot be a zero-length string")

        output = "IDX KIND POS_X POS_Y POS_Z EUL_PSI EUL_THETA EUL_PHI"
        output = output.replace(" ", sep)
        text = [comment + output] + [
            sep.join([str(idx), voxel.to_text(sep=sep)])
            for idx, voxel in enumerate(self.fractal)
        ]
        return "\n".join(text)

    def to_frame(self, suppress_hash: bool = False):
        """Convert voxelised representation to data frame

        :param suppress_hash: Hide the hash in front of the 'IDX' column
            which is kept for compatibiilty with the Geant4
            DNA simulation format

        :returns: Voxelised fractal as a data frame

        """
        rows = []
        hash_option = "" if suppress_hash else "#"
        for idx, voxel in enumerate(self.fractal):
            ss = voxel.to_series()
            ss[hash_option + "IDX"] = idx
            rows.append(ss)
        df = pd.DataFrame(rows)
        return df[
            [
                hash_option + "IDX",
                "KIND",
                "POS_X",
                "POS_Y",
                "POS_Z",
                "EUL_PSI",
                "EUL_THETA",
                "EUL_PHI",
            ]
        ]

    def to_plot(self, refine: int = 0, batch: bool = False) -> plt.figure:
        """
        Create a matplotlib figure instance of this fractal

        fig = toPlot(refine=0, batch=False)
        :param refine: points to plot in between voxels (more points = clearer path)
        :param batch: True to suppress automatic display of the figure
        """
        pts = [vox.pos for vox in self.fractal]

        refinedpts = []
        for ii in range(0, len(pts) - 1):
            refinedpts.append(pts[ii])
            step = (pts[ii + 1] - pts[ii]) / (refine + 1)
            for jj in range(1, refine + 1):
                refinedpts.append(pts[ii] + step * jj)
        refinedpts.append(pts[len(pts) - 1])

        pts = np.array(refinedpts)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.plot(pts[:, 0], pts[:, 1], pts[:, 2])

        if batch is not True:
            fig.show()

        return fig

    def to_pretty_plot(
        self, refine: int = 10, mayavi: bool = False, mask: Callable = None
    ):
        """Create a matplotlib figure instance of this fractal, with curved lines
        rather than hard corners

        fig = to_pretty_plot(refine=10, batch=False)

        :param refine: points to plot in between voxels (more points = clearer path)
        :param mayavi: make plot with MayaVI
        :param mask: Callable function that returns true for voxels that should be
            plotted. mask is a function that should take a position three vector and
            return boolean True/False for any position.
        """
        if mask is not None:
            if not callable(mask):
                raise ValueError("Mask should be callable")
        pts = [vox.pos for vox in self.fractal]

        # replace pts with an array containing the sides of boxes rather than
        # the centres of boxes. We are then going to interp curves b/w points
        midpoints = [0.5 * (ii + jj) for ii, jj in zip(pts[0:-1], pts[1:])]

        refinedpts = [pts[0]]
        for ii in range(0, len(midpoints) - 1):
            # print(ii)
            step = 1 / (refine)
            entry_point = midpoints[ii]
            exit_point = midpoints[ii + 1]
            entry_normal = midpoints[ii] - pts[ii]
            exit_normal = midpoints[ii + 1] - pts[ii + 1]
            interp = self._interpolator(
                entry_point, entry_normal, exit_point, exit_normal
            )
            for jj in range(0, refine):
                # print(jj*step)
                refinedpts.append(interp(step * jj))
        refinedpts.append(pts[-1])

        pts = np.array(refinedpts)
        idx = np.round(np.arange(len(pts)) / len(pts), 3)
        pts = np.concatenate([pts, idx.reshape([len(idx), 1])], axis=1)
        if mayavi:
            if mask is not None:
                # iterate over plot_points to find acceptable points
                plot_points = [ii for (ii, pos) in enumerate(pts) if mask(pos)]
                grouped_plot_points = []
                while len(plot_points) > 0:
                    end_point = plot_points.pop()
                    grouped_plot_points
                    current_point = end_point
                    while plot_points[-1] == current_point - 1:
                        current_point = plot_points.pop()
                        if len(plot_points) == 0:
                            break
                    grouped_plot_points.append((current_point, end_point))
                pts = [pts[start : end + 1] for start, end in grouped_plot_points]

            else:
                pts = [pts]
            fig = mlab.figure(bgcolor=(1.0, 1.0, 1.0))
            for arr in pts:
                assert np.all(arr[:, 3] >= 0) and np.all(
                    arr[:, 3] <= 1
                ), """color value was {}, outside acceptable range
                    [0, 1]""".format(
                    arr[:, 3]
                )
                mlab.plot3d(
                    arr[:, 0],
                    arr[:, 1],
                    arr[:, 2],
                    arr[:, 3],
                    # color=(0., .8, 0),
                    colormap="Spectral",
                    tube_radius=0.1,
                    vmin=0.0,
                    vmax=1.0,
                )
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.plot(pts[:, 0], pts[:, 1], pts[:, 2])

        return fig

    def center_fractal(self) -> None:
        """Center the fractal around (x, y, z) = (0, 0, 0)"""
        minvals = np.array([np.inf, np.inf, np.inf])
        maxvals = np.array([-np.inf, -np.inf, -np.inf])

        # identify max/min values
        for voxel in self.fractal:
            for (ii, (minv, v, maxv)) in enumerate(zip(minvals, voxel.pos, maxvals)):
                if v < minv:
                    minvals[ii] = v
                elif v > maxv:
                    maxvals[ii] = v

        # transform
        transform = -minvals - (maxvals - minvals) / 2.0
        for voxel in self.fractal:
            oldpos = voxel.pos
            voxel.pos = oldpos + transform

        return None

    def _interpolator(self, point_entry, norm_entry, point_exit, norm_exit):
        """Interpolator for pretty_plot"""
        # 1. case 1, norm_entry = norm_exit
        if (norm_entry == norm_exit).all():
            interp = lambda x: point_entry + x * (point_exit - point_entry)
            return interp
        # 2. Case 2, circular interpolation
        # 2a. find centre of circle
        # pdb.set_trace()
        norm_plane = np.cross(norm_entry, norm_exit)
        d1 = -np.dot(point_entry, norm_entry)
        d2 = -np.dot(point_exit, norm_exit)
        d3 = -np.dot(point_entry, norm_plane)
        centre = -(
            d1 * np.cross(norm_exit, norm_plane)
            + d2 * np.cross(norm_plane, norm_entry)
            + d3 * np.cross(norm_entry, norm_exit)
        ) / (np.dot(norm_entry, np.cross(norm_exit, norm_plane)))
        v_init = point_entry - centre
        v_final = point_exit - centre
        rotation_axis = np.cross(v_init, v_final)
        rotation_axis /= np.linalg.norm(rotation_axis)  # unit vector

        # function to change the magnitude of the vector
        mag = lambda x: np.linalg.norm(v_init) + x * (
            np.linalg.norm(v_final) - np.linalg.norm(v_init)
        )

        vec = lambda x: np.dot(rot_ax_angle(rotation_axis, x * np.pi / 2.0), v_init)

        return lambda x: centre + 2 * mag(x) * vec(x)

    @staticmethod
    def _makeVoxel(prevVoxel: Voxel, currpos: np.array, nextpos: np.array) -> Voxel:
        """Make the next voxel in a chain.

        :param prevVoxel: Previous Voxel in Chain
        :param currpos: Position (centre) of the voxel to make
        :param nexpos: Position (centre) of the next voxel to make
        """
        # clean and vet output
        currpos = np.around(currpos, 8)
        nextpos = np.around(nextpos, 8)
        prevpos = prevVoxel.pos
        firstChange = currpos - prevpos
        secondChange = nextpos - currpos
        perp = np.cross(firstChange, secondChange)
        turn = perp.any()
        if turn:
            perp = perp.round()
            return Voxel(
                currpos,
                prevVoxel.outHeading,
                prevVoxel.outPrincipal,
                secondChange,
                perp,
            )
        else:
            return Voxel(
                currpos,
                prevVoxel.outHeading,
                prevVoxel.outPrincipal,
                prevVoxel.outHeading,
                prevVoxel.outPrincipal,
            )

    @classmethod
    def fromSeed(cls, seed: str, iterations: int, distance: float = 1):
        """Make a voxelised fractal from an L-String seed.

        The available L-Strings are described in fractaldna.structure_models.hilbert

        :param seed: Seed L-String
        :param iterations: number of times to iterate seed
        :param distance: distance between voxels

        :returns: Voxelised Fractal representation
        """
        for n in range(iterations):
            seed = hilbert.iterate_lstring(seed)
        return cls.fromLString(seed, distance=distance)

    @classmethod
    def from_path(
        cls, path: Union[List[np.array], List[List], np.array], pbar: bool = False
    ):
        """Convert a path (a list of xyz positions) to a voxelised representation

        :param path: path to convert
        :param pbar: show a progress bar if true

        :returns: Voxelised Fractal representation
        """
        path = np.asarray(path)
        if path.shape[1] != 3:
            raise ValueError("Positions in path require 3 elements (xyz)")

        # Check that we only ever move in the six principal directions
        for this_pos, next_pos in zip(path, np.roll(path, -1, axis=0)[:-1]):
            absdiff = np.abs(this_pos - next_pos)
            if np.sum(absdiff) != 1:
                raise ValueError(
                    f"Going from {this_pos} to {next_pos} is not valid. "
                    "Points should move by one unit in one direction only."
                )

        lastpos = np.array([2 * path[len(path) - 1] - path[len(path) - 2]])
        path = np.concatenate([path, lastpos])

        vf = cls()

        arrpath = np.array(path)

        mins = np.zeros(3)  # array [xmin, ymin, zmin]
        maxs = np.zeros(3)  # array [xmax, ymax, zmax]
        lens = np.zeros(3)

        for ii in range(3):
            mins[ii] = min(arrpath[:, ii])
            maxs[ii] = max(arrpath[:, ii])
            lens[ii] = maxs[ii] - mins[ii]

        first = arrpath[0]
        second = arrpath[1]
        zeroheading = second - first
        zeroposition = first - zeroheading
        zeroprincipal = np.array([1, 0, 0])
        if (np.around(zeroheading, 8) == zeroprincipal).all():
            zeroprincipal = np.array([0, 1, 0])
        zeroVoxel = Voxel(
            zeroposition, zeroheading, zeroprincipal, zeroheading, zeroprincipal
        )

        vf.fractal = [cls._makeVoxel(zeroVoxel, first, second)]

        stop = len(path) - 1
        iters = tqdm(
            range(1, stop), disable=(not pbar), total=stop - 1, desc="Building Voxels"
        )
        for ii in iters:
            vf.fractal.append(
                cls._makeVoxel(vf.fractal[ii - 1], arrpath[ii], arrpath[ii + 1])
            )
            # print np.around(arrpath[ii], 3)
        # print "Path Length: ", len(vf.fractal)

        return vf

    @classmethod
    def fromLString(cls, lstring: str, pbar: bool = False):
        """Convert an L-String into a VoxelisedFractal

        :param lstring: L-String to convert
        :param pbar: Show a progress bar for generating path and voxels

        :returns: Voxelised Fractal representation
        """
        path = hilbert.generate_path(lstring, n=1, distance=1, pbar=pbar)

        return cls.from_path(path, pbar=pbar)



import logging
import os

logger = logging.getLogger("fractaldna")

log_level = os.environ.get("FRACTALDNA_LOG_LEVEL", "ERROR")

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(log_level)

# create formatter
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# add formatter to ch
ch.setFormatter(formatter)



from typing import List

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D  # NOQA
from tqdm import tqdm

# interpret F as DrawForward(1);
# interpret + as Yaw(90);
# interpret - as Yaw(-90);
# interpret n as Pitch(90);
# interpret & as Pitch(-90);
# interpret > as Roll(90);
# interpret < as Roll(-90);
# interpret | as Yaw(180);

A = r"B-F+CFC+F-D&FnD-F+&&CFC+F+B<<"
B = r"A&FnCFBnFnDnn-F-Dn|FnB|FCnFnA<<"
C = r"|Dn|FnB-F+CnFnA&&FA&FnC+F+BnFnD<<"
D = r"|CFB-F+B|FA&FnA&&FB-F+B|FC<<"

X = r"n<XFn<XFX-Fn>>XFX&F+>>XFX-F>X->"


# Legacy Peano Curves
# This Peano works infinitely - maybe not! but goes to at least n=3 iterations
P = r"P>>FP>>FP>>+F+P>>FP>>FP>>+F+P>>FP>>FP>>&F&P>>FP>>FP>>+F+P>>FP>>FP>>+F+P>>FP>>FP>>&F&P>>FP>>FP>>+F+P>>FP>>FP>>+F+P>>FP>>FP"  # NOQA

# # These Peanos do not work
# # Y = r"FF-F-FF+F+FFnFn+FF-F-FF+F+FF&F&+FF-F-FF+F+FF"  # Peano bottom to top
# T = r"TFUFT+F+UFTFU-F-TFUFTnFn-ZFYFZ+F+YFZFY-F-ZFYFZ&F&-TFUFT+F+UFTFU-F-TFUFT"
# U = T[::-1]
# Y = r"YFZFY-F-ZFYFZ+F+YFZFYnFn+UFTFU-F-TFUFT+F+UFTFU&F&+YFZFY-F-ZFYFZ+F+YFZFY"
# Z = Y[::-1]

# # This guy needs some testing
# # R = r"RFRFR-F-RFRFR+F+RFRFRnFn+RFRFR-F-RFRFR+F+RFRFR&F&+RFRFR-F-RFRFR+F+RFRFR<<"  # NOQA
# R = r"R>>FR>>FR>>+F+R>>FR>>FR>>+F+R>>FR>>FRnFn>>-R>>FR>>FR>>-F-R>>FR>>FR>>-F-R>>FR>>FRnFn>>+R>>FR>>FR>>+F+R>>FR>>FR>>+F+R>>FR>>FR"  # NOQA

SUBSTITUTIONS = {
    "A": A,
    "B": B,
    "C": C,
    "D": D,
    "X": X,
    "P": P
    # "P": P,
    # "Y": Y,
    # "Z": Z,
    # "T": T,
    # "U": U,
    # "R": R,
}


def rotz(angle):
    return np.array(
        [
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1],
        ]
    )


def roty(angle):
    return np.array(
        [
            [np.cos(angle), 0, np.sin(angle)],
            [0, 1, 0],
            [-np.sin(angle), 0, np.cos(angle)],
        ]
    )


def rotx(angle):
    return np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)],
        ]
    )


# Each of these is a rotation in the frame of the local axis
# The local axis is rotated from the global axis (Identity)
# Ax_local = R*Ax_global  ==> R = Ax_local*Ax_global_inv = Ax_local
# RotH is a rotation about the local X axis
# RotH = R*RotX = Ax_local*RotX


def roth(local_axis, angle):
    return np.dot(local_axis, np.dot(rotx(angle), np.linalg.inv(local_axis)))


def rotl(local_axis, angle):
    return np.dot(local_axis, np.dot(roty(angle), np.linalg.inv(local_axis)))


def rotu(local_axis, angle):
    return np.dot(local_axis, np.dot(rotz(angle), np.linalg.inv(local_axis)))


def iterate_lstring(inString: str):
    """
    Iterate an L-String one time

    :param inString:  L-string to iterate
    :return: Iterated L-String
    """
    outString = []
    append = outString.append
    for char in inString:
        if char in SUBSTITUTIONS:
            append(SUBSTITUTIONS[char])
        else:
            append(char)

    return "".join(outString)


def generate_path(
    lstring: str,
    n: int = 2,
    distance: float = 10.0,
    rounding: int = 0,
    pbar: bool = False,
) -> List[np.array]:
    """
    Generate a path from an l-string

    When moving a distance smaller than 1, ensure that rounding is set correctly.

    This algorithm works best when moving in integer steps and applying any scaling
    after generating the path.

    :param lstring: lstring describing path
    :param n: steps on path between forward movements
    :param distance: distance between points forward movements
    :param rounding: rounding to apply to each position
    :param pbar: display a progress bar
    :return: list of XYZ points
    """
    if distance < 10 ** (-1 * rounding):
        raise ValueError(
            f"Decimal place rounding is too large {rounding} places "
            f"for the distance moved forward {distance} units"
        )

    axis = np.eye(3)
    pos = [np.array([0, 0, 0])]

    # Fuction to move a distance in a set direction
    # taking n steps
    def forward(axis, n=n, distance=distance):
        heading = np.dot(axis, np.array([1, 0, 0]))
        return [(ii + 1) * distance * heading / n for ii in range(0, n)]

    charFunctions = {
        r"+": lambda axis: np.dot(rotu(axis, +np.pi / 2.0), axis),
        r"-": lambda axis: np.dot(rotu(axis, -np.pi / 2.0), axis),
        r"&": lambda axis: np.dot(rotl(axis, +np.pi / 2.0), axis),
        r"n": lambda axis: np.dot(rotl(axis, -np.pi / 2.0), axis),
        r">": lambda axis: np.dot(roth(axis, +np.pi / 2.0), axis),
        r"<": lambda axis: np.dot(roth(axis, -np.pi / 2.0), axis),
        r"|": lambda axis: np.dot(rotu(axis, np.pi), axis),
    }

    pbar_lstring = tqdm(
        lstring, total=len(lstring), disable=(not pbar), desc="Generating Path"
    )
    for char in pbar_lstring:
        if char == r"F":
            lastpos = pos[len(pos) - 1]
            for point in forward(axis):
                if rounding is None:
                    pos.append(lastpos + point)
                else:
                    pos.append(np.round(lastpos + point, decimals=rounding))
        elif char in charFunctions.keys():
            axis = charFunctions[char](axis)
        else:
            pass

    return pos



#!/bin/python
# Small script to generate isotropically distributed non-overlapping prisms
# Used with molecular-dna in Geant4 to generate test geometries
#
# For usage, type python prisms.py help
#
# (c) Nathanael Lampe, 2016

from typing import List, Set, Union

import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import Axes3D, proj3d  # NOQA
from tqdm import tqdm


class Prism(object):
    """ """

    def __init__(
        self, center: np.array, size: np.array, axis: np.array, rotation: float = 0
    ):
        """
        Prism(center, size, axis, rotation=0)

        Make a prism object
        x-axis is translated to point along the axis vector.
        rotation is the rotation of the shape around the x-axis
        """
        assert center.shape == (3,), "center should be a 3-vector"
        assert size.shape == (3,), "size should be a 3-vector"
        assert axis.shape == (3,), "axis should be a 3-vector"
        self.center = center
        self.size = size
        self.axis = axis / np.linalg.norm(axis)
        self.angx = rotation
        self.angy = -np.arctan2(axis[2], (axis[0] ** 2 + axis[1] ** 2) ** 0.5)
        self.angz = np.arctan2(axis[1], axis[0])
        rx = np.array(
            [
                [1, 0, 0],
                [0, np.cos(self.angx), -np.sin(self.angx)],
                [0, np.sin(self.angx), np.cos(self.angx)],
            ]
        )
        ry = np.array(
            [
                [np.cos(self.angy), 0, np.sin(self.angy)],
                [0, 1, 0],
                [-np.sin(self.angy), 0, np.cos(self.angy)],
            ]
        )
        rz = np.array(
            [
                [np.cos(self.angz), -np.sin(self.angz), 0],
                [np.sin(self.angz), np.cos(self.angz), 0],
                [0, 0, 1],
            ]
        )
        self.rotation = np.dot(rz, np.dot(ry, rx))
        self.norm1 = np.dot(self.rotation, np.array([0, 0, 1]))
        self.norm2 = np.dot(self.rotation, np.array([0, -1, 0]))

        corners = [
            0.5 * np.array([size[0], size[1], size[2]]),
            0.5 * np.array([size[0], size[1], -size[2]]),
            0.5 * np.array([size[0], -size[1], size[2]]),
            0.5 * np.array([size[0], -size[1], -size[2]]),
            0.5 * np.array([-size[0], size[1], size[2]]),
            0.5 * np.array([-size[0], size[1], -size[2]]),
            0.5 * np.array([-size[0], -size[1], size[2]]),
            0.5 * np.array([-size[0], -size[1], -size[2]]),
        ]
        self.corners = [
            self.center + np.dot(self.rotation, corner) for corner in corners
        ]

    def to_text(self) -> str:
        """
        Describe prism according to the following text specification:
        POS_X POS_Y POS_Z EUL_PSI EUL_THETA EUL_PHI
        """
        euler = get_euler_angles(self.rotation)
        return " ".join(
            map(
                str,
                [
                    self.center[0],
                    self.center[1],
                    self.center[2],
                    euler[0],
                    euler[1],
                    euler[2],
                ],
            )
        )

    def to_series(self) -> pd.Series:
        """Return the prism as a pandas series object."""
        euler = get_euler_angles(self.rotation)
        ss = pd.Series(
            {
                "TYPE": "prism",
                "POS_X": self.center[0],
                "POS_Y": self.center[1],
                "POS_Z": self.center[2],
                "EUL_PSI": euler[0],
                "EUL_THETA": euler[1],
                "EUL_PHI": euler[2],
            }
        )
        return ss

    def to_plot(self, n: int = 200, ax: plt.Axes = None, arrows: bool = False):
        """
        Prism.to_plot(n)

        plot in 3d using n points
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
        x = []
        y = []
        z = []
        for ii in range(n):
            p = self.get_point()
            x.append(p[0])
            y.append(p[1])
            z.append(p[2])
        ax.scatter(x, y, z)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")

        if arrows is True:
            for (axis, sz) in zip([self.axis, self.norm1, self.norm2], self.size):
                a = Arrow3D(
                    self.center,
                    axis * sz + self.center,
                    mutation_scale=20,
                    lw=1,
                    arrowstyle="-|>",
                    color="k",
                )
                ax.add_artist(a)

        if ax is None:
            return fig
        else:
            return None

    def does_overlap(self, other):
        """
        Evaluate if the two prisms overlap using a separating axis test
        This is reasonably easy with prisms, as we only need to compare to
        the axes of each wall, using all the corners.
        """
        assert isinstance(other, type(self)), "argument must be a object"

        d = sum((self.center - other.center) ** 2)
        sum_rads = sum(self.size**2) + sum(other.size**2)

        if d > sum_rads:
            return False
        else:
            # do a separating axis test
            test_axes = [
                self.axis,
                self.norm1,
                self.norm2,
                other.axis,
                other.norm1,
                other.norm2,
            ]

            overlaps = [
                is_separating_axis(ax, self.corners, other.corners) for ax in test_axes
            ]

            separating_axis_exists = True in overlaps

            return not separating_axis_exists

    def get_point(self):
        """
        Return a point inside the prism
        """
        random = np.random.rand(3) - 0.5
        point = self.center + np.dot(self.rotation, random * self.size)
        return point

    def contains_point(self, point):
        """
        Prism.contains_point(point)

        Returns True if point is in prism
        """
        overlaps = [
            is_separating_axis(ax, self.corners, point)
            for ax in [self.axis, self.norm1, self.norm2]
        ]
        return True not in overlaps


class PrismList(object):
    """
    A List of Non-Overlapping Rectangular Prisms

    This class facilitates generation of a set of prisms so that when a new
    prism is added, it:

    a) does not overlap other prisms and
    b) only checks for overlaps against nearby prisms

    A 'prism grid' is constructed which divides the space within
    '+/- extent' of the grid centre into cells with side length
    specified by 'resolution'. This helper grid is used so that
    the routine only needs to see if the prism being placed overlaps
    local prisms.


    :param center: XYZ-position where the prism should be centered
    :param extent: XYZ-half lengths for the maximum bounding cube
                   of all prisms to be placed.
    :param resolution: The size of a 'cell' in the prism grid.

    """

    def __init__(self, center: List, extent: List, resolution: float):
        if len(center) != 3:
            raise ValueError("center must be a 3-vector")
        if len(extent) != 3:
            raise ValueError("extent must be a 3-vector")

        self.center = np.array(center)
        self.extent = np.array(extent)
        self.resolution = resolution

        # Calculate the maximum number of divisions needed along the
        # longest axis of the prism.
        self.divisions = 2 * max(self.extent) // self.resolution + 1

        # Make a list to hold lists of all prisms potentially in a grid cell.
        # If we denote a cell in the x direction by i, a cell in the y direction
        # by j, and a cell in the z-direction by k, the unique index in the
        # prism grid of the ijk-th cell is int(i + j*divisions + k*divisions^2)
        self.prism_grid = [[] for ii in range(int(self.divisions**3))]
        self.prisms = []

    def append(self, prism: Prism) -> bool:
        """Append a new prism to the Prism list

        :param prism: prism to be appended.
        :returns: True if a prism is placed, else False
        """
        if not isinstance(prism, Prism):
            raise ValueError("Expected a Prism object")

        # get prism grid cells to check
        indices = self.get_prism_grid_indices(prism)

        for index in indices:
            for test_prism in self.prism_grid[index]:
                if prism.does_overlap(test_prism):
                    return False

        # if no overlaps, append the prism to the prism grid
        for index in indices:
            self.prism_grid[index].append(prism)
        self.prisms.append(prism)

        return True

    def to_frame(self, suppress_hash: bool = False) -> pd.DataFrame:
        """Generate a Data Frame describing the prisms

        :param suppress_hash: Hide the hash in front of the 'IDX' column
            which is kept for compatibiilty with the Geant4
            DNA simulation format

        :returns: Voxelised fractal as a data frame
        """
        series_list = []
        hash_option = "" if suppress_hash else "#"
        for ii, prism in enumerate(self.prisms):
            ss = prism.to_series()
            ss[hash_option + "IDX"] = ii
            series_list.append(ss)
        df = pd.DataFrame(series_list)
        df = df[
            [
                hash_option + "IDX",
                "TYPE",
                "POS_X",
                "POS_Y",
                "POS_Z",
                "EUL_PSI",
                "EUL_THETA",
                "EUL_PHI",
            ]
        ]
        return df

    def get_prism_grid_indices(self, prism: Prism) -> Set[int]:
        """Get the prism grid indices"""
        # get the corners of the incoming prism
        ijks = [self._position_to_ijk(corner) for corner in prism.corners]
        ijks = np.array(ijks)

        # Get the x, y and z co-ordinates spanned by the prism
        ii = np.arange(min(ijks[:, 0]), max(ijks[:, 0] + 1), 1)
        jj = np.arange(min(ijks[:, 1]), max(ijks[:, 1] + 1), 1)
        kk = np.arange(min(ijks[:, 2]), max(ijks[:, 2] + 1), 1)
        indices = set(
            [
                self._ijk_to_index(np.array([i_, j_, k_]))
                for i_ in ii
                for j_ in jj
                for k_ in kk
            ]
        )
        return indices

    def _position_to_ijk(self, position: np.array) -> np.array:
        """Convert a position to ijk-coordinates in the prism grid"""
        # transform the local position onto the grid (which is +ve values only)
        local = position - self.center + self.extent
        ijk = local // self.resolution
        return ijk

    def _index_to_ijk(self, index: int) -> np.array:
        """Convert an index value in the prism grid back to ijk coords

        This reverses the equation:
        index = int(i + j*divisions + k*divisions^2)

        where i,j,k < divisions
        """
        ii = index % self.divisions
        jj = ((index - ii) % self.divisions**2) / self.divisions
        kk = (index - ii - self.divisions * jj) / self.divisions**2
        return np.array([ii, jj, kk])

    def _ijk_to_index(self, ijk: Union[np.array, List]) -> int:
        """Convert an ijk reference in the 'prism grid' to a single index:

        The ijk-th cell is in the grid is int(i + j*divisions + k*divisions^2)
        """
        return int(
            ijk[0] + self.divisions * ijk[1] + self.divisions**2 * ijk[2]
        )  # NOQA


def get_euler_angles(rotmatrix: np.array) -> np.array:
    """Get Euler Angles from Rotation Matrix

    :param rotmatrix: Rotation Matrix
    :returns: Euler Angles
    """
    sintheta = rotmatrix[2, 0]
    if abs(sintheta) != 1:
        theta = -np.arcsin(rotmatrix[2, 0])
        costheta = np.cos(theta)
        psi = np.arctan2(rotmatrix[2, 1] / costheta, rotmatrix[2, 2] / costheta)
        phi = np.arctan2(rotmatrix[1, 0] / costheta, rotmatrix[0, 0] / costheta)
    else:
        phi = 0
        if sintheta < 0:  # Positive case
            theta = np.pi / 2.0
            psi = phi + np.arctan2(rotmatrix[0, 1], rotmatrix[0, 2])
        else:
            theta = -np.pi / 2.0
            psi = -phi + np.arctan2(-rotmatrix[0, 1], -rotmatrix[0, 2])

    return np.array([psi, theta, phi])


class Arrow3D(FancyArrowPatch):
    """Helper class to plot arrows"""

    def __init__(self, start, end, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = ([start[0], end[0]], [start[1], end[1]], [start[2], end[2]])

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


def random_point_in_ball(rad=1) -> np.array:
    """
    random_point_in_ball(rad=1)

    Return a point uniformly distributed in the radius 'rad' ball

    :param rad: Radius of ball
    :returns: 3-vector array
    """
    p = 2 * (np.random.rand(3) - 0.5)
    while sum(p**2) > 1:
        p = 2 * (np.random.rand(3) - 0.5)
    return rad * p


def random_direction() -> np.array:
    """
    Generate a random 3-vector

    :returns: euler angles for a random direction
    """
    theta = 2 * np.pi * np.random.rand()
    u = 2 * (np.random.rand() - 0.5)
    k = (1 - u * u) ** 0.5
    return np.array([k * np.cos(theta), k * np.sin(theta), u])


def is_separating_axis(axis, points1, points2):
    """
    test_projection(axis, points1, points2)

    Test if two sets of points in 3-D overlap or not after projection onto an
    axis
    """
    if type(points1) != list:
        points1 = [points1]
    if type(points2) != list:
        points2 = [points2]
    distances1 = np.array([np.dot(axis, point) for point in points1])
    distances2 = np.array([np.dot(axis, point) for point in points2])
    grouped = [(distance, distances2) for distance in distances1]

    d1_smaller = [np.all(d < dists2) for (d, dists2) in grouped]
    d1_bigger = [np.all(d > dists2) for (d, dists2) in grouped]

    is_d1_smaller = False if False in d1_smaller else True
    is_d1_bigger = False if False in d1_bigger else True

    can_be_separated = is_d1_smaller or is_d1_bigger
    # return True if there is an overlap (both are false)
    return can_be_separated


def generate_non_overlapping_prisms(
    n_prisms: int,
    size: Union[np.array, List],
    rad: float,
    early_exit: int = -1,
    verbose: bool = False,
) -> PrismList:
    """Generate a collection of non-overlapping prisms

    Candidate prisms are generated at random and placed in a spherical volume.
    Overlaps are then checked and a new candidate is chosen if there is an overlap.
    The algorithm will exit early if early_exit is set to a number greater than 0.

    :param n_prisms: Number of prisms to generate
    :param size: size of prism in x, y, z directions
    :param rad: radius of ball in which to generate prisms
    :param early_exit: stop attempting to place 'early_exit' prisms due to too many overlaps (-1 to disable)
    :param verbose: Display progress of algorithm
    """
    size = np.asarray(size)
    if size.shape != (3,):
        raise ValueError(f"Size should be a 3-vector like np.array([1, 2, 3])")

    def new_prism(rad=rad):
        position = random_point_in_ball(rad)
        axis = random_direction()
        rotation = np.random.rand() * 2 * np.pi
        return Prism(position, size, axis, rotation)

    mag = np.sqrt(np.sum(size * size))
    extent = np.ones(3) * (mag + rad)
    n_prisms_placed = 0
    prisms = PrismList(np.zeros(3), extent, (mag + rad) / 6)
    prisms.append(new_prism(rad))
    attempts = 0
    if verbose is True:
        pbar = tqdm(iterable=None, total=n_prisms)
    while n_prisms_placed < n_prisms:
        if attempts >= early_exit and early_exit >= 0:
            sys.stderr.write(
                f"Too many failed prism placements, stopped after {n_prisms_placed} placements."
            )
            break
        this_prism = new_prism()

        if prisms.append(this_prism) is True:
            if verbose is True:
                pbar.update(1)
            attempts = 0
            n_prisms_placed += 1
        else:
            attempts += 1
    if verbose is True:
        pbar.close()
    return prisms


# if __name__ == "__main__":
#     usage = """
#     prisms.py num N size X Y Z rad R
#     num  N: number of prisms to generate
#     size X Y Z: dimensions of prisms to place
#     rad R: radius of sphere in which to place prisms
#     """
#     if "help" in sys.argv:
#         print(usage)
#         sys.exit(0)

#     err1 = (len(sys .argv) != 9)
#     err2 = ("num" not in sys.argv) or ("size" not in sys.argv) or\
#            ("rad" not in sys.argv)

#     if err1 or err2:
#         print(usage)
#     else:
#         try:
#             n_idx = sys.argv.index("num")
#             n = int(sys.argv[n_idx + 1])
#             size_idx = sys.argv.index("size")
#             sx = float(sys.argv[size_idx + 1])
#             sy = float(sys.argv[size_idx + 2])
#             sz = float(sys.argv[size_idx + 3])
#             size = np.array([sx, sy, sz])
#             rad_idx = sys.argv.index("rad")
#             rad = float(sys.argv[rad_idx + 1])
#         except:
#             print(usage)
#             sys.exit(0)

#         prisms = non_overlapping_prisms(n, size, rad, verbose=False)
#         header = "# IDX TYPE POS_X POS_Y POS_Z EUL_PSI EUL_THETA EUL_PHI\n"
#         outstring = "\n".join(["{} prism ".format(ii) + p.to_text()
#                                for ii, p in enumerate(prisms.prisms)])
#         sys.stdout.write(header + outstring)



"""
dnapositions.py

Specification of DNA molecules in a variety of co-ordinate systems
"""

from typing import Dict, List

import pdb
import re
from copy import deepcopy
from itertools import combinations

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from fractaldna.utils.logging import logger

# Start by specifying the molecules in cylindrical co-ords
# From Arnott and Hukins, 1972, Biochem and Biophys Res Comms, 47, 6, p1504.

# B-DNA geometry
# Each array contains r (angstroms), theta (deg), z (angstroms) position
PHOSPHATE = {
    "O1": np.array([8.75, 97.4, 3.63]),
    "O2": np.array([10.20, 91.1, 1.86]),
    "O3": np.array([8.82, 103.3, 1.29]),
    "P1": np.array([8.91, 95.2, 2.08]),
    "O4": np.array([7.73, 88.0, 1.83]),
}

DEOXYRIBOSE = {
    "C5": np.array([7.70, 79.8, 2.77]),
    "O5": np.array([6.22, 66.0, 1.83]),
    "C4": np.array([7.59, 69.9, 2.04]),
    "C3": np.array([8.20, 69.9, 0.64]),
    "C2": np.array([7.04, 73.2, -0.24]),
    "C1": np.array([5.86, 67.4, 0.47]),
}

ADENINE = {
    "N9": np.array([4.63, 76.6, 0.42]),
    "C8": np.array([4.84, 93.0, 0.50]),
    "N7": np.array([3.95, 105.4, 0.43]),
    "C5": np.array([2.74, 94.0, 0.28]),
    "N6": np.array([1.83, 154.0, 0.14]),
    "C6": np.array([1.41, 107.2, 0.15]),
    "N1": np.array([0.86, 40.1, 0.03]),
    "C2": np.array([2.17, 30.6, 0.04]),
    "N3": np.array([3.24, 47.0, 0.16]),
    "C4": np.array([3.33, 70.5, 0.28]),
}

GUANINE = {
    "N9": np.array([4.63, 76.6, 0.42]),
    "C8": np.array([4.82, 93.2, 0.50]),
    "N7": np.array([3.92, 105.7, 0.42]),
    "C5": np.array([2.70, 94.0, 0.28]),
    "O6": np.array([1.71, 154.6, 0.13]),
    "C6": np.array([1.39, 109.3, 0.15]),
    "N1": np.array([0.92, 37.9, 0.03]),
    "N2": np.array([3.01, 4.2, -0.10]),
    "C2": np.array([2.28, 28.7, 0.03]),
    "N3": np.array([3.29, 46.7, 0.16]),
    "C4": np.array([3.33, 70.3, 0.28]),
}

CYTOSINE = {
    "N1": np.array([4.63, 76.6, 0.42]),
    "C6": np.array([4.99, 92.2, 0.52]),
    "C5": np.array([4.35, 107.0, 0.47]),
    "N4": np.array([2.76, 136.6, 0.27]),
    "C4": np.array([2.94, 110.0, 0.32]),
    "N3": np.array([2.31, 83.9, 0.22]),
    "O2": np.array([3.69, 47.9, 0.18]),
    "C2": np.array([3.40, 67.4, 0.27]),
}

THYMINE = {
    "N1": np.array([4.63, 76.6, 0.42]),
    "C6": np.array([5.01, 92.3, 0.52]),
    "Me": np.array([5.40, 119.8, 0.58]),
    "C5": np.array([4.38, 106.9, 0.47]),
    "O4": np.array([2.82, 136.3, 0.27]),
    "C4": np.array([2.98, 111.9, 0.32]),
    "N3": np.array([2.36, 85.2, 0.23]),
    "O2": np.array([3.64, 47.8, 0.18]),
    "C2": np.array([3.42, 67.3, 0.27]),
}

# Entry of these numbers double-checked 16-MAR-2016 by NJL

# Van der Waals radius of varius elements in Angstrom.
# From Bondi (1964), J. Phys. Chem.; and
# Kammeyer & Whitman (1972), J. Chem. Phys.
RADIUS = {"H": 1.2, "C": 1.7, "O": 1.4, "N": 1.5, "P": 1.9, "Me": 2.1}

COLORS = {
    "H": "white",
    "C": "grey",
    "O": "red",
    "N": "skyblue",
    "P": "goldenrod",
    "Me": "grey",
    "PHOSPHATE": "yellow",
    "DEOXYRIBOSE": "black",
    "ADENINE": "orange",
    "GUANINE": "green",
    "CYTOSINE": "red",
    "THYMINE": "blue",
}

LETTERS = re.compile("[A-Za-z]+")


def opposite_pair(base: Dict[str, np.array]):
    if base == THYMINE:
        return ADENINE
    elif base == ADENINE:
        return THYMINE
    elif base == GUANINE:
        return CYTOSINE
    elif base == CYTOSINE:
        return GUANINE
    else:
        return None


def base_name(base: Dict[str, np.array]):
    if base == THYMINE:
        return "THYMINE"
    elif base == ADENINE:
        return "ADENINE"
    elif base == GUANINE:
        return "GUANINE"
    elif base == CYTOSINE:
        return "CYTOSINE"
    else:
        return None


def overlap_volume(pos1: List, pos2: List, r1: float, r2: float) -> float:
    """
    Calculate overlapping volume of two spheres

    overlap_volume(pos1, pos2, r1, r2)

    :param pos1: Position of first molecule (x, y, z)
    :param pos2: Position of second molecule (x, y, z)
    :param r1: Radius of first molecule
    :param r2: Radius of second molecule
    :return: Volume of overlap.
    """
    d = sum((pos1 - pos2) ** 2) ** 0.5
    # check they overlap
    if d >= (r1 + r2):
        return 0
    # check if one entirely holds the other
    if r1 > (d + r2):  # 2 is entirely contained in one
        return 4.0 / 3.0 * np.pi * r2**3
    if r2 > (d + r1):  # 1 is entirely contained in one
        return 4.0 / 3.0 * np.pi * r1**3

    vol = (
        np.pi
        * (r1 + r2 - d) ** 2
        * (d**2 + (2 * d * r1 - 3 * r1**2 + 2 * d * r2 - 3 * r2**2) + 6 * r1 * r2)
    ) / (12 * d)
    return vol


def get_p_values(a, b, c, alpha, beta, gamma):
    """
    Helper function for triple_overlap_volume
    """
    t2 = (a + b + c) * (-a + b + c) * (a - b + c) * (a + b - c)
    tabg2 = (
        (a + beta + gamma)
        * (-a + beta + gamma)
        * (a - beta + gamma)
        * (a + beta - gamma)
    )

    t = t2**0.5
    tabg = tabg2**0.5

    a2 = a**2
    b2 = b**2
    c2 = c**2

    alpha2 = alpha**2
    beta2 = beta**2
    gamma2 = gamma**2

    p1 = ((b2 - c2 + beta2 - gamma2) ** 2 + (t - tabg) ** 2) / (4 * a2) - alpha2  # NOQA
    p2 = ((b2 - c2 + beta2 - gamma2) ** 2 + (t + tabg) ** 2) / (4 * a2) - alpha2  # NOQA

    return p1, p2


def atanpi(val):
    theta = np.arctan(val)
    if theta < 0:
        theta += np.pi
    return theta


def triple_overlap_volume(pos1, pos2, pos3, r1, r2, r3):
    """
    triple_overlap_volume(pos1, pos2, pos3, r1, r2, r3)

    Calculate volume overlapped by 3 spheres
    From Gibson and Scheraga (1987)

    Note:
    There are cases where this formula doesn't work properly, not documented
    in the paper by Gibson and Scheraga.This corresponds to the case where
    the center of one circle lies between the line joining the point of
    intersection of the three cricles, and the line between the center of the
    two other circles.

    This geometry rarely arises for chemical species, but if negative volumes
    start appearing, or other quantities that seem unlikely, a Monte Carlo
    integration can be used (provided in this package).
    """
    a = sum((pos3 - pos2) ** 2) ** 0.5
    b = sum((pos3 - pos1) ** 2) ** 0.5
    c = sum((pos2 - pos1) ** 2) ** 0.5
    if not ((a <= (r3 + r2)) and (b <= (r3 + r1)) and (c <= (r2 + r1))):
        return 0

    # Check if one sphere entirely contains another
    if (r1 > (b + r3)) or (r1 > (c + r2)):  # Circle 1 encloses circle 2/3
        vol = overlap_volume(pos2, pos3, r2, r3)
        return vol
    elif (r2 > (a + r3)) or (r2 > (c + r1)):  # Circle 2 encloses circle 1/3
        vol = overlap_volume(pos1, pos3, r1, r3)
        return vol
    elif (r3 > (b + r1)) or (r3 > (a + r2)):  # Circle 3 encloses circle 1/2
        vol = overlap_volume(pos1, pos2, r1, r2)
        return vol

    if (r1 > b) and (r2 > a):  # Circle C is enclosed by both others
        logger.debug("Warning:: Circle C's center is interior to A and B")
    if (r1 > c) and (r3 > a):  # Circle B is enclosed by both others
        logger.debug("Warning:: Circle B's center is interior to A and C")
    if (r2 > c) and (r3 > b):  # Circle A is enclosed by both others
        logger.debug("Warning:: Circle A's center is interior to B and C")

    alpha = r1
    beta = r2
    gamma = r3

    a2 = a**2
    b2 = b**2
    c2 = c**2

    alpha2 = alpha**2
    beta2 = beta**2
    gamma2 = gamma**2

    eps1 = (beta2 - gamma2) / a2
    eps2 = (gamma2 - alpha2) / b2
    eps3 = (alpha2 - beta2) / c2

    w2 = (
        (alpha2 * a2 + beta2 * b2 + gamma2 * c2) * (a2 + b2 + c2)
        - 2 * (alpha2 * a2**2 + beta2 * b2**2 + gamma2 * c2**2)
        + (a2 * b2 * c2) * (eps1 * eps2 + eps2 * eps3 + eps3 * eps1 - 1)
    )

    if w2 > 0:
        w = w2**0.5
        q1 = a * (b2 + c2 - a2 + beta2 + gamma2 - 2.0 * alpha2 + eps1 * (b2 - c2))
        q2 = b * (c2 + a2 - b2 + gamma2 + alpha2 - 2.0 * beta2 + eps2 * (c2 - a2))
        q3 = c * (a2 + b2 - c2 + alpha2 + beta2 - 2.0 * gamma2 + eps3 * (a2 - b2))

        alpha3 = alpha**3.0
        beta3 = beta**3.0
        gamma3 = gamma**3.0
        aw = a * w
        bw = b * w
        cw = c * w

        vol = (
            w / 6.0
            - a
            / 2.0
            * (beta2 + gamma2 - a2 * (1.0 / 6.0 - eps1**2 / 2.0))
            * atanpi(2 * w / q1)  # NOQA
            - b
            / 2.0
            * (gamma2 + alpha2 - b2 * (1.0 / 6.0 - eps2**2 / 2.0))
            * atanpi(2 * w / q2)  # NOQA
            - c
            / 2.0
            * (alpha2 + beta2 - c2 * (1.0 / 6.0 - eps3**2 / 2.0))
            * atanpi(2 * w / q3)  # NOQA
            + (2.0 / 3.0)
            * alpha3
            * (
                atanpi(bw / (alpha * q2) * (1 - eps2))
                + atanpi(cw / (alpha * q3) * (1 + eps3))
            )  # NOQA
            + (2.0 / 3.0)
            * beta3
            * (
                atanpi(cw / (beta * q3) * (1 - eps3))
                + atanpi(aw / (beta * q1) * (1 + eps1))
            )  # NOQA
            + (2.0 / 3.0)
            * gamma3
            * (
                atanpi(aw / (gamma * q1) * (1 - eps1))
                + atanpi(bw / (gamma * q2) * (1 + eps2))
            )
        )  # NOQA

    elif w2 < 0:
        p1, p2 = get_p_values(a, b, c, alpha, beta, gamma)
        p3, p4 = get_p_values(b, c, a, beta, gamma, alpha)
        p5, p6 = get_p_values(c, a, b, gamma, alpha, beta)

        if (p3 > 0) and (p5 > 0):
            if p1 > 0:
                vol = 0
            if p1 < 0:
                vol = overlap_volume(pos2, pos3, r2, r3)
        elif (p1 > 0) and (p5 > 0):  # fill out...
            if p3 > 0:
                vol = 0
            if p3 < 0:
                vol = overlap_volume(pos1, pos3, r1, r3)
        elif (p1 > 0) and (p3 > 0):
            if p5 > 0:
                vol = 0
            if p5 < 0:
                vol = overlap_volume(pos1, pos2, r1, r2)
        elif (p1 > 0) and (p3 < 0) and (p5 < 0):  # NOQA
            vol = (
                overlap_volume(pos1, pos2, r1, r2)
                + overlap_volume(pos1, pos3, r1, r3)
                - 4.0 / 3.0 * np.pi * r1**3.0
            )
        elif (p1 < 0) and (p3 > 0) and (p5 < 0):  # NOQA
            vol = (
                overlap_volume(pos1, pos2, r1, r2)
                + overlap_volume(pos2, pos3, r2, r3)
                - 4.0 / 3.0 * np.pi * r2**3.0
            )
        elif (p1 < 0) and (p3 < 0) and (p5 > 0):  # NOQA
            vol = (
                overlap_volume(pos1, pos3, r1, r3)
                + overlap_volume(pos2, pos3, r2, r3)
                - 4.0 / 3.0 * np.pi * r3**3.0
            )
        else:
            # Fall back to MCMC calculation
            vol = mc_triple_volume(pos1, pos2, pos3, r1, r2, r3)
    else:
        vol = 0

    return vol


def mc_triple_volume(p1, p2, p3, r1, r2, r3, n=1e5):
    # Generate points inside the box containing smallest circle
    # as this is a constraint
    rs = [r1, r2, r3]
    if r1 == min(rs):
        centres = p1
        ranges = 2 * np.ones([3]) * r1
    elif r2 == min(rs):
        centres = p2
        ranges = 2 * np.ones([3]) * r2
    elif r3 == min(rs):
        centres = p3
        ranges = 2 * np.ones([3]) * r3

    in_overlap = 0.0
    in_circle = lambda p, c, r: sum((p - c) ** 2.0) ** 0.5 < r
    for ii in range(int(n)):
        position = (np.random.random(3) - 0.5) * ranges + centres
        in1 = in_circle(position, p1, r1)
        in2 = in_circle(position, p2, r2)
        in3 = in_circle(position, p3, r3)
        if in1 and in2 and in3:
            in_overlap += 1
    vol_total = np.product(ranges)
    return vol_total * in_overlap / n


class MoleculeFromAtoms:
    def __init__(self, atoms: Dict[str, np.array]):
        """
        MoleculeFromAtoms(atoms)
        A molecule created from a dictionary of cartesian atom positions
        """
        self.atoms = deepcopy(atoms)

    @classmethod
    def from_cylindrical(cls, atoms: Dict[str, np.array], inverse: bool = False):
        """Make a MoleculeFromAtoms instance from a list of atoms in cylindrical
        coords (r, theta, phi)

        MoleculeFromAtoms.from_cylindrical(atoms, inverse=False)

        Note:
        Theta is in degrees

        :param inverse: (default False)Set to True to generate a dyadically related
            base pair (negates theta and z)
        """
        cylindrical = deepcopy(atoms)
        cartesian = {}
        sgn = 1 if inverse is False else -1
        for (name, pos) in cylindrical.items():
            z = sgn * pos[2]  # NOQA
            y = pos[0] * np.sin(sgn * np.pi * pos[1] / 180.0)
            x = pos[0] * np.cos(sgn * np.pi * pos[1] / 180.0)
            cartesian[name] = np.array([x, y, z])

        return cls(cartesian)

    def find_center(self) -> float:
        """
        c = MoleculeFromAtoms.find_center()

        Find the barycenter of the atoms that constitute this molecule
        """
        c = np.zeros([3])
        denom = 0
        for (atom, pos) in self.atoms.items():
            r = RADIUS[LETTERS.match(atom).group()]
            c += pos * r
            denom += r
        return c / denom

    def find_half_lengths(self) -> np.array:
        """
        half_lengths = MoleculeFromAtoms.find_half_lengths()

        Return an array of the half lengths of a box that encloses the molecule
        in xyz. This is not a minimum bounding volume
        """
        c = self.find_center()
        extents = []
        for (atom, pos) in self.atoms.items():
            pos -= c
            rad = RADIUS[LETTERS.match(atom).group()]
            x = (pos[0] - rad) if pos[0] <= 0 else (pos[0] + rad)
            y = (pos[1] - rad) if pos[1] <= 0 else (pos[1] + rad)
            z = (pos[2] - rad) if pos[2] <= 0 else (pos[2] + rad)
            extents.append([x, y, z])

        extents = np.array(extents)
        max_extents = np.max(extents, axis=0)
        min_extents = np.min(extents, axis=0)
        return 0.5 * (max_extents - min_extents)

    def find_equivalent_half_lengths(self) -> np.array:
        """
        l = MoleculeFromAtoms.find_equivalent_half_lengths()

        Find the half lengths scaled to give a volume equal to what the
        constituent molecules occupy
        """
        half_lengths = self.find_half_lengths()
        max_volume = 4.0 / 3.0 * np.pi * np.product(half_lengths)
        equiv_volume = 4.0 / 3.0 * np.pi * self.find_equivalent_radius() ** 3.0
        return half_lengths * (equiv_volume / max_volume) ** (1.0 / 3.0)

    def find_radius(self) -> float:
        """
        r = MoleculeFromAtoms.find_radius()

        Return the minimum radius that encloses this molecule
        """
        c = self.find_center()
        radii = []
        for (atom, pos) in self.atoms.items():
            pos -= c
            rad = RADIUS[LETTERS.match(atom).group()]
            rad += np.sqrt(np.sum(pos * pos))
            radii.append(rad)

        return np.max(radii)

    def find_equivalent_radius(self) -> float:
        """
        r = MoleculeFromAtoms.find_equivalent_radius()

        Return the radius that yields the same volume occupied by all atoms
        """
        vol = 0
        for (atom, pos) in self.atoms.items():
            rad = RADIUS[LETTERS.match(atom).group()]
            vol += 4.0 / 3.0 * np.pi * rad**3
        # subtract double overlaps
        for ((a1, p1), (a2, p2)) in combinations(self.atoms.items(), 2):
            r1 = RADIUS[LETTERS.match(a1).group()]
            r2 = RADIUS[LETTERS.match(a2).group()]
            vol -= overlap_volume(p1, p2, r1, r2)
        for ((a1, p1), (a2, p2), (a3, p3)) in combinations(self.atoms.items(), 3):
            # pdb.set_trace()
            r1 = RADIUS[LETTERS.match(a1).group()]
            r2 = RADIUS[LETTERS.match(a2).group()]
            r3 = RADIUS[LETTERS.match(a3).group()]
            vol += triple_overlap_volume(p1, p2, p3, r1, r2, r3)
        return (vol * 3.0 / 4.0 / np.pi) ** (1.0 / 3.0)

    def to_plot(self) -> plt.Figure:
        """
        fig = MoleculeFromAtoms.to_plot()

        Returns a matplotlib figure instance of the molecule.
        """
        atomsets = {}
        for (atom, pos) in self.atoms.items():
            a = LETTERS.match(atom).group()
            if a not in atomsets:
                rad = RADIUS[a] if a in RADIUS else 2
                col = COLORS[a] if a in COLORS else "blue"
                atomsets[a] = {"radius": rad, "positions": [], "color": col}
            atomsets[a]["positions"].append(pos)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")
        for atomset in atomsets.values():
            pos = np.array(atomset["positions"])
            size = 3.14 * atomset["radius"] * atomset["radius"]
            color = atomset["color"]
            ax.scatter(pos[:, 0], pos[:, 1], pos[:, 2], s=20 * size, c=color)

        return fig

    def __len__(self) -> int:
        """
        len(MoleculeFromAtoms)
        returns number of atoms
        """
        return self.atoms.__len__()


class MoleculeDictionary:
    """
    Base class for test molecules
    """

    def __init__(self):
        self.atoms = {}

    def items(self):
        return self.atoms.items()


class DoubleStrand(MoleculeDictionary):
    """
    Double strand of DNA (for testing)
    """

    def __init__(self):
        super().__init__()
        sequence = [THYMINE, GUANINE, ADENINE, CYTOSINE, THYMINE] * 2
        for ii in range(len(sequence)):
            for (k, v) in sequence[ii].items():
                x = v[0]
                y = v[1] + 36 * ii
                z = v[2] + 3.4 * ii
                self.atoms[k + "_LEFTBP" + str(ii)] = np.array([x, y, z])

            for (k, v) in DEOXYRIBOSE.items():
                x = v[0]
                y = v[1] + 36 * ii
                z = v[2] + 3.4 * ii
                self.atoms[k + "_LEFTDO" + str(ii)] = np.array([x, y, z])

            for (k, v) in PHOSPHATE.items():
                x = v[0]
                y = v[1] + 36 * ii
                z = v[2] + 3.4 * ii
                self.atoms[k + "_LEFTPO" + str(ii)] = np.array([x, y, z])

            for (k, v) in opposite_pair(sequence[ii]).items():
                x = v[0]
                y = -v[1] + 36 * ii
                z = -v[2] + 3.4 * ii
                self.atoms[k + "_RIGHTBP" + str(ii)] = np.array([x, y, z])

            for (k, v) in DEOXYRIBOSE.items():
                x = v[0]
                y = -v[1] + 36 * ii
                z = -v[2] + 3.4 * ii
                self.atoms[k + "_RIGHTDO" + str(ii)] = np.array([x, y, z])

            for (k, v) in PHOSPHATE.items():
                x = v[0]
                y = -v[1] + 36 * ii
                z = -v[2] + 3.4 * ii
                self.atoms[k + "_RIGHTPO" + str(ii)] = np.array([x, y, z])

        return None


class DoubleStrandMolecules(MoleculeDictionary):
    """
    Double strand of DNA (for testing)
    """

    def __init__(self):
        super().__init__()
        sequence = [THYMINE, GUANINE, ADENINE, CYTOSINE, THYMINE] * 2
        for ii in range(len(sequence)):
            # left
            bp = sequence[ii]
            name = base_name(bp)

            c = MoleculeFromAtoms(bp).find_center()
            self.atoms[name + "_LEFT" + str(ii)] = self.shift(c, ii)
            c = MoleculeFromAtoms(DEOXYRIBOSE).find_center()
            self.atoms["DEOXYRIBOSE_LEFT" + str(ii)] = self.shift(c, ii)
            c = MoleculeFromAtoms(PHOSPHATE).find_center()
            self.atoms["PHOSPHATE_LEFT" + str(ii)] = self.shift(c, ii)

            # right
            bp = opposite_pair(bp)
            name = base_name(bp)
            c = MoleculeFromAtoms(bp).find_center()
            self.atoms[name + "_RIGHT" + str(ii)] = self.shift(c, ii, inv=True)
            c = MoleculeFromAtoms(DEOXYRIBOSE).find_center()
            self.atoms["DEOXYRIBOSE_RIGHT" + str(ii)] = self.shift(c, ii, inv=True)
            c = MoleculeFromAtoms(PHOSPHATE).find_center()
            self.atoms["PHOSPHATE_RIGHT" + str(ii)] = self.shift(c, ii, inv=True)

        return None

    @staticmethod
    def shift(v, ii, inv=False):
        sign = -1 if inv is True else 1
        x = v[0]
        y = sign * v[1] + 36 * ii
        z = sign * v[2] + 3.4 * ii
        return np.array([x, y, z])



# type: ignore[attr-defined]
"""FractalDNA is a Python package built to generate DNA geometries for simulations"""

import sys

try:
    from importlib import metadata as importlib_metadata
except ImportError:  # for Python<3.8
    import importlib_metadata


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fractaldna',
 'fractaldna.dna_models',
 'fractaldna.structure_models',
 'fractaldna.utils']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=4.8.2,<5.0.0',
 'matplotlib>=3.4.3,<4.0.0',
 'mistune==0.8.4',
 'numpy>=1.22.0,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'pytest>=6.2.5,<8.0.0',
 'rich>=10.7,<13.0',
 'scipy>=1.8.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'typer[all]>=0.3.2,<0.5.0']

entry_points = \
{'console_scripts': ['fractaldna = fractaldna.__main__:app']}

setup_kwargs = {
    'name': 'fractaldna',
    'version': '0.6.0',
    'description': 'FractalDNA is a Python package built to generate DNA geometries for simulations',
    'long_description': 'FractalDNA\n===\nPython routines for generating geometric models of DNA\n---\n\n*FractalDNA is being converted to a package, it is under active developmemt*\n\n<div align="center">\n\n[![Build status](https://github.com/natl/fractaldna/workflows/build/badge.svg?branch=master&event=push)](https://github.com/natl/fractaldna/actions?query=workflow%3Abuild)\n[![Python Version](https://img.shields.io/pypi/pyversions/fractaldna.svg)](https://pypi.org/project/fractaldna/)\n[![Dependencies Status](https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg)](https://github.com/natl/fractaldna/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/natl/fractaldna/blob/master/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--versions-e10079.svg)](https://github.com/natl/fractaldna/releases)\n[![License](https://img.shields.io/github/license/natl/fractaldna)](https://github.com/natl/fractaldna/blob/master/LICENSE)\n\n</div>\n\nFractalDNA is a Python package to make DNA geometries that can be joined together like\njigsaw puzzles. Both simple, sections of DNA and Solenoidal DNA can be built. This\nmodule was built to enable DNA-level simulations to be run in [Geant4-DNA](http://geant4-dna.in2p3.fr/), part of the\n[Geant4](geant4.cern.ch/) project.\n\nStructure models define the large scale structure of DNA,\nseeded from fractals. An example seeding fractal is below:\n\n<p align="center">\n  <img width="460" height="300" src="https://raw.githubusercontent.com/natl/fractaldna/master/docs/source/images/fractal-path.svg" alt="A 3-D iterated Hilbert Curve">\n</p>\n\nDNA Models provide straight and curved segments that can come together to\nmake DNA for use in simulations.\n\n<p align="center">\n  <img width="460" height="300" src="https://raw.githubusercontent.com/natl/fractaldna/master/docs/source/images/single_solenoid_line_plot.jpg" alt="A straight solenoidal DNA segment">\n</p>\n\nProject documentation is available [here](http://natl.github.io/fractaldna/) alongside [notebook examples](http://natl.github.io/fractaldna/examples.html)\n\n## ⚙️ Installation\n\nInstall FractalDNA with `pip`\n\n```bash\npip install fractaldna\n```\n\nor install with `Poetry`\n\n```bash\npoetry add fractaldna\n```\n\n## 🧬 Make some DNA\n\n```py\nfrom fractaldna.dna_models import dnachain as dna\n\n# Make a DNA Chain 40 base pairs long (repeating GATC).\nchain = dna.DNAChain("GTAC" * 10)\n\n# Export it to a DataFrame for use in another program\ndf = chain.to_frame()\n```\n\nFor more, check out the [notebook examples](http://natl.github.io/fractaldna/examples.html) in the project documentation.\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/natl/fractaldna)](https://github.com/natl/fractaldna/blob/master/LICENSE)\n\nThis project is licensed under the terms of the `MIT` license. See [LICENSE](https://github.com/natl/fractaldna/blob/master/LICENSE) for more details.\n\n## 📃 Citation\n\n```bibtex\n@misc{fractaldna,\n  author = {Nathanael Lampe},\n  title = {FractalDNA},\n  year = {2021},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/natl/fractaldna}}\n}\n```\n\n## Credits [![🚀 Your next Python package needs a bleeding-edge project structure.](https://img.shields.io/badge/python--package--template-%F0%9F%9A%80-brightgreen)](https://github.com/TezRomacH/python-package-template)\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template)',
    'author': 'Nathanael Lampe',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/natl/fractaldna',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
